module Main where
import Data.Maybe
import qualified Data.List as List
import qualified Data.Map.Strict as Map
import qualified Data.Set as Set
import Lexer
import Parser
import Grammar

-- Hypothesis
maybeHypothesis context expression = case Map.lookup expression context of
                                       Nothing -> MyNothing expression
                                       Just a  -> Hypothesis a expression

isHypothesis context expression = (not . isMyNothing) (maybeHypothesis context expression)

-- Axiom
maybeAxiom expr@(Impl a (Impl b a'))                                         | a == a'                                    = Axiom 1 expr
maybeAxiom expr@(Impl (Impl a b) (Impl (Impl a' (Impl b' c)) (Impl a'' c'))) | a == a' && a' == a'' && b == b' && c == c' = Axiom 2 expr
maybeAxiom expr@(Impl a (Impl b (And a' b')))                                | a == a' && b == b'                         = Axiom 3 expr
maybeAxiom expr@(Impl (And a b) a')                                          | a == a'                                    = Axiom 4 expr
maybeAxiom expr@(Impl (And a b) b')                                          | b == b'                                    = Axiom 5 expr
maybeAxiom expr@(Impl a (Or a' b))                                           | a == a'                                    = Axiom 6 expr
maybeAxiom expr@(Impl b (Or a b'))                                           | b == b'                                    = Axiom 7 expr
maybeAxiom expr@(Impl (Impl a c) (Impl (Impl b c') (Impl (Or a' b') c'')))   | a == a' && b == b' && c == c' && c' == c'' = Axiom 8 expr
maybeAxiom expr@(Impl (Impl a b) (Impl (Impl a' (Not b')) (Not a'')))        | a == a' && a' == a'' && b == b'            = Axiom 9 expr
maybeAxiom expr@(Impl (Not (Not a)) a')                                      | a == a'                                    = Axiom 10 expr
maybeAxiom expr                                                                                                           = MyNothing expr

isAxiom expression = (not . isMyNothing) (maybeAxiom expression)

-- Modus Ponens
maybeModusPonens to !implMap !allAsMap =
                                    let
                                    fromList' = (Map.lookup to implMap)
                                    fromList = if isJust fromList' then fromJust fromList' else []
                                    left' = List.find (\l -> isJust (Map.lookup l allAsMap)) fromList
                                    leftIndex  = if isJust left' then fromJust (Map.lookup (fromJust left') allAsMap) else -1
                                        in
                                            if leftIndex /= -1 then
                                                let
                                                implIndex = fromJust (Map.lookup (Impl (fromJust left') to) allAsMap)
                                                    in
                                                        ModusPonens (fromJust left') to
                                            else
                                                MyNothing to

isModusPonens to !implMap !allAsMap = (not . isMyNothing) (maybeModusPonens to implMap allAsMap)

-- Check other conditions
isMyNothing smth = case smth of
     (MyNothing a) -> True
     _             -> False

isImpl expr = case expr of
     (Impl a b) -> True
     _          -> False

-- Output
showProblem []      result = "|- " ++ (show result) ++ "\n"
showProblem [last] result = (show last) ++ " " ++ (showProblem [] result)
showProblem (x:xs) result = (show x) ++ ", " ++ (showProblem xs result)

writeAxiom10 expression = do
    putStrLn $ show (Impl expression (Impl (Not (Not expression)) expression))
    putStrLn $ show (Impl (Impl expression (Impl (Not (Not expression)) expression)) (Impl (Not (Impl (Not (Not expression)) expression)) (Impl expression (Impl (Not (Not expression)) expression))))
    putStrLn $ show (Impl (Not (Impl (Not (Not expression)) expression)) (Impl expression (Impl (Not (Not expression)) expression)))
    putStrLn $ show (Impl (Not (Impl (Not (Not expression)) expression)) (Impl expression (Not (Impl (Not (Not expression)) expression))))
    putStrLn $ show (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Not (Impl (Not (Not expression)) expression)) (Not (Impl (Not (Not expression)) expression))))
    putStrLn $ show (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Impl (Not (Impl (Not (Not expression)) expression)) (Not (Impl (Not (Not expression)) expression))) (Not (Impl (Not (Not expression)) expression))))
    putStrLn $ show (Impl (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Not (Impl (Not (Not expression)) expression)) (Not (Impl (Not (Not expression)) expression)))) (Impl (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Impl (Not (Impl (Not (Not expression)) expression)) (Not (Impl (Not (Not expression)) expression))) (Not (Impl (Not (Not expression)) expression)))) (Impl (Not (Impl (Not (Not expression)) expression)) (Not (Impl (Not (Not expression)) expression)))))
    putStrLn $ show (Impl (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Impl (Not (Impl (Not (Not expression)) expression)) (Not (Impl (Not (Not expression)) expression))) (Not (Impl (Not (Not expression)) expression)))) (Impl (Not (Impl (Not (Not expression)) expression)) (Not (Impl (Not (Not expression)) expression))))
    putStrLn $ show (Impl (Not (Impl (Not (Not expression)) expression)) (Not (Impl (Not (Not expression)) expression)))
    putStrLn $ show (Impl (Impl expression (Impl (Not (Not expression)) expression)) (Impl (Impl expression (Not (Impl (Not (Not expression)) expression))) (Not expression)))
    putStrLn $ show (Impl (Impl (Impl expression (Impl (Not (Not expression)) expression)) (Impl (Impl expression (Not (Impl (Not (Not expression)) expression))) (Not expression))) (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Impl expression (Impl (Not (Not expression)) expression)) (Impl (Impl expression (Not (Impl (Not (Not expression)) expression))) (Not expression)))))
    putStrLn $ show (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Impl expression (Impl (Not (Not expression)) expression)) (Impl (Impl expression (Not (Impl (Not (Not expression)) expression))) (Not expression))))
    putStrLn $ show (Impl (Impl (Not (Impl (Not (Not expression)) expression)) (Impl expression (Impl (Not (Not expression)) expression))) (Impl (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Impl expression (Impl (Not (Not expression)) expression)) (Impl (Impl expression (Not (Impl (Not (Not expression)) expression))) (Not expression)))) (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Impl expression (Not (Impl (Not (Not expression)) expression))) (Not expression)))))
    putStrLn $ show (Impl (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Impl expression (Impl (Not (Not expression)) expression)) (Impl (Impl expression (Not (Impl (Not (Not expression)) expression))) (Not expression)))) (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Impl expression (Not (Impl (Not (Not expression)) expression))) (Not expression))))
    putStrLn $ show (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Impl expression (Not (Impl (Not (Not expression)) expression))) (Not expression)))
    putStrLn $ show (Impl (Impl (Not (Impl (Not (Not expression)) expression)) (Impl expression (Not (Impl (Not (Not expression)) expression)))) (Impl (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Impl expression (Not (Impl (Not (Not expression)) expression))) (Not expression))) (Impl (Not (Impl (Not (Not expression)) expression)) (Not expression))))
    putStrLn $ show (Impl (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Impl expression (Not (Impl (Not (Not expression)) expression))) (Not expression))) (Impl (Not (Impl (Not (Not expression)) expression)) (Not expression)))
    putStrLn $ show (Impl (Not (Impl (Not (Not expression)) expression)) (Not expression))
    putStrLn $ show (Impl (Not expression) (Impl (Not (Not expression)) expression))
    putStrLn $ show (Impl (Impl (Not expression) (Impl (Not (Not expression)) expression)) (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Not expression) (Impl (Not (Not expression)) expression))))
    putStrLn $ show (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Not expression) (Impl (Not (Not expression)) expression)))
    putStrLn $ show (Impl (Impl (Not (Impl (Not (Not expression)) expression)) (Not expression)) (Impl (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Not expression) (Impl (Not (Not expression)) expression))) (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Not (Not expression)) expression))))
    putStrLn $ show (Impl (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Not expression) (Impl (Not (Not expression)) expression))) (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Not (Not expression)) expression)))
    putStrLn $ show (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Not (Not expression)) expression))
    putStrLn $ show (Impl (Impl (Not (Impl (Not (Not expression)) expression)) (Impl (Not (Not expression)) expression)) (Impl (Impl (Not (Impl (Not (Not expression)) expression)) (Not (Impl (Not (Not expression)) expression))) (Not (Not (Impl (Not (Not expression)) expression)))))
    putStrLn $ show (Impl (Impl (Not (Impl (Not (Not expression)) expression)) (Not (Impl (Not (Not expression)) expression))) (Not (Not (Impl (Not (Not expression)) expression))))
    putStrLn $ show (Not (Not (Impl (Not (Not expression)) expression)))

writeAxiomAndHypothesis expression = do
    putStrLn $ show expression
    putStrLn $ show (Impl expression (Impl (Not expression) expression))
    putStrLn $ show (Impl (Not expression) expression)
    putStrLn $ show (Impl (Not expression) (Impl (Not expression) (Not expression)))
    putStrLn $ show (Impl (Impl (Not expression) (Impl (Not expression) (Not expression))) (Impl (Impl (Not expression) (Impl (Impl (Not expression) (Not expression)) (Not expression))) (Impl (Not expression) (Not expression))))
    putStrLn $ show (Impl (Impl (Not expression) (Impl (Impl (Not expression) (Not expression)) (Not expression))) (Impl (Not expression) (Not expression)))
    putStrLn $ show (Impl (Not expression) (Impl (Impl (Not expression) (Not expression)) (Not expression)))
    putStrLn $ show (Impl (Not expression) (Not expression))
    putStrLn $ show (Impl (Impl (Not expression) expression) (Impl (Impl (Not expression) (Not expression)) (Not (Not expression))))
    putStrLn $ show (Impl (Impl (Not expression) (Not expression)) (Not (Not expression)))
    putStrLn $ show (Not (Not expression))

writeModusPonens from to = do
    putStrLn $ show (Impl (Impl (Not to) (Not (Impl from to))) (Impl (Impl (Not to) (Not (Not (Impl from to)))) (Not (Not to))))
    putStrLn $ show (Impl (Not (Not (Impl from to))) (Impl (Not to) (Not (Not (Impl from to)))))
    putStrLn $ show (Not (Not (Impl from to)))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Not to)))
    putStrLn $ show (Impl (Impl (Impl from to) (Not to)) (Impl (Impl (Impl from to) (Not (Not to))) (Not (Impl from to))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Not to)) (Impl (Impl (Impl from to) (Not (Not to))) (Not (Impl from to)))) (Impl (Not to) (Impl (Impl (Impl from to) (Not to)) (Impl (Impl (Impl from to) (Not (Not to))) (Not (Impl from to))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Not to)) (Impl (Impl (Impl from to) (Not (Not to))) (Not (Impl from to)))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Not to))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Not to)) (Impl (Impl (Impl from to) (Not (Not to))) (Not (Impl from to))))) (Impl (Not to) (Impl (Impl (Impl from to) (Not (Not to))) (Not (Impl from to))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Not to)) (Impl (Impl (Impl from to) (Not (Not to))) (Not (Impl from to))))) (Impl (Not to) (Impl (Impl (Impl from to) (Not (Not to))) (Not (Impl from to)))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Not (Not to))) (Not (Impl from to))))
    putStrLn $ show (Impl (Impl from to) (Impl (Impl from to) (Impl from to)))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl from to))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl from to)))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl from to))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl from to)) (Impl from to))) (Impl (Impl from to) (Impl from to))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl from to)) (Impl from to))) (Impl (Impl from to) (Impl from to)))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl from to)) (Impl from to))) (Impl (Impl from to) (Impl from to))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl from to)) (Impl from to))) (Impl (Impl from to) (Impl from to)))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl from to)))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl from to)) (Impl from to))) (Impl (Impl from to) (Impl from to))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl from to)) (Impl from to))) (Impl (Impl from to) (Impl from to))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl from to)) (Impl from to))) (Impl (Impl from to) (Impl from to))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl from to)) (Impl from to))) (Impl (Impl from to) (Impl from to)))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl from to)) (Impl from to))) (Impl (Impl from to) (Impl from to))))
    putStrLn $ show (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl from to)) (Impl from to)))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl from to)) (Impl from to))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl from to)) (Impl from to)))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl from to)) (Impl from to))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl from to)) (Impl from to)))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl from to)) (Impl from to))) (Impl (Impl from to) (Impl from to)))) (Impl (Not to) (Impl (Impl from to) (Impl from to)))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl from to)) (Impl from to))) (Impl (Impl from to) (Impl from to)))) (Impl (Not to) (Impl (Impl from to) (Impl from to))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl from to)))
    putStrLn $ show (Impl (Impl from to) (Impl (Not to) (Impl from to)))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Not to) (Impl (Impl from to) (Impl (Not to) (Impl from to)))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Not to) (Impl from to))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from to)))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from to))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from to)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from to))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Not to) (Impl from to)))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from to)))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from to)))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from to)))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from to))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from to)))))
    putStrLn $ show (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))
    putStrLn $ show (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Not to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))
    putStrLn $ show (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Not to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from to)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from to)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from to)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from to)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from to))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from to)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from to)))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from to))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from to)) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))))
    putStrLn $ show (Impl (Not to) (Impl from (Not to)))
    putStrLn $ show (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Not to) (Impl (Not to) (Impl from (Not to)))))
    putStrLn $ show (Impl (Not to) (Impl (Not to) (Impl from (Not to))))
    putStrLn $ show (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))
    putStrLn $ show (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))) (Impl (Not to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Not to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))) (Impl (Not to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))) (Impl (Not to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))
    putStrLn $ show (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))))) (Impl (Not to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))
    putStrLn $ show (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Not to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))
    putStrLn $ show (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Not to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))) (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))) (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl from (Not to))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl from (Not to)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl from (Not to))) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))))) (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Not to) (Impl (Impl from (Not to)) (Not from)))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from)))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from)))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Impl (Impl from (Not to)) (Not from))) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from)))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from)))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl from to)) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl from to)) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl from to)) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl from to)) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl from to))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl from to)) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl from to)) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from)))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Not to) (Impl (Impl from to) (Impl (Not to) (Not from))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl from to) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Not to) (Not from))))) (Impl (Not to) (Impl (Impl from to) (Impl (Not to) (Not from)))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Not to) (Not from))))
    putStrLn $ show (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from))))
    putStrLn $ show (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Not to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))))
    putStrLn $ show (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))))) (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from))))))
    putStrLn $ show (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))
    putStrLn $ show (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))
    putStrLn $ show (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))
    putStrLn $ show (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))
    putStrLn $ show (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))
    putStrLn $ show (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Not to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))
    putStrLn $ show (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not from)))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not from))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))))
    putStrLn $ show (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))
    putStrLn $ show (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Not to) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))
    putStrLn $ show (Impl (Not to) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))
    putStrLn $ show (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl from to) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))
    putStrLn $ show (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl from to) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))) (Impl (Not to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl from to) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl from to) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Not to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl from to) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl from to) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))
    putStrLn $ show (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))
    putStrLn $ show (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))) (Impl (Not to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))))
    putStrLn $ show (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))))) (Impl (Not to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))))
    putStrLn $ show (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))
    putStrLn $ show (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Not to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))
    putStrLn $ show (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Not to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))
    putStrLn $ show (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Not to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Not to) (Not (Not from))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Not to) (Not (Not from)))) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))))
    putStrLn $ show (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to))))) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Impl (Not to) (Not from)) (Impl (Impl (Not (Not from)) (Impl (Impl (Not to) (Not (Not from))) (Not (Not to)))) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))))
    putStrLn $ show (Impl (Impl (Impl from to) (Impl (Not to) (Not from))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to))))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Impl (Not to) (Not from))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Not from))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to))))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Not from))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to)))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Not to) (Not from)))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Not from))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to))))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not to) (Not from))) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to))))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Not to) (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to)))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Impl (Not to) (Not from)) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to)))))) (Impl (Not to) (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to)))))
    putStrLn $ show (Not (Not from))
    putStrLn $ show (Impl (Not to) (Not (Not from)))
    putStrLn $ show (Impl (Not (Not from)) (Impl (Impl from to) (Not (Not from))))
    putStrLn $ show (Impl (Impl (Not (Not from)) (Impl (Impl from to) (Not (Not from)))) (Impl (Not to) (Impl (Not (Not from)) (Impl (Impl from to) (Not (Not from))))))
    putStrLn $ show (Impl (Not to) (Impl (Not (Not from)) (Impl (Impl from to) (Not (Not from)))))
    putStrLn $ show (Impl (Impl (Not to) (Not (Not from))) (Impl (Impl (Not to) (Impl (Not (Not from)) (Impl (Impl from to) (Not (Not from))))) (Impl (Not to) (Impl (Impl from to) (Not (Not from))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Not (Not from)) (Impl (Impl from to) (Not (Not from))))) (Impl (Not to) (Impl (Impl from to) (Not (Not from)))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Not (Not from))))
    putStrLn $ show (Impl (Impl (Impl from to) (Not (Not from))) (Impl (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to)))) (Impl (Impl from to) (Not (Not to)))))
    putStrLn $ show (Impl (Impl (Impl (Impl from to) (Not (Not from))) (Impl (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to)))) (Impl (Impl from to) (Not (Not to))))) (Impl (Not to) (Impl (Impl (Impl from to) (Not (Not from))) (Impl (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to)))) (Impl (Impl from to) (Not (Not to)))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Not (Not from))) (Impl (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to)))) (Impl (Impl from to) (Not (Not to))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Not (Not from)))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Not (Not from))) (Impl (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to)))) (Impl (Impl from to) (Not (Not to)))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to)))) (Impl (Impl from to) (Not (Not to)))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Not (Not from))) (Impl (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to)))) (Impl (Impl from to) (Not (Not to)))))) (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to)))) (Impl (Impl from to) (Not (Not to))))))
    putStrLn $ show (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to)))) (Impl (Impl from to) (Not (Not to)))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to))))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to)))) (Impl (Impl from to) (Not (Not to))))) (Impl (Not to) (Impl (Impl from to) (Not (Not to))))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Impl (Not (Not from)) (Not (Not to)))) (Impl (Impl from to) (Not (Not to))))) (Impl (Not to) (Impl (Impl from to) (Not (Not to)))))
    putStrLn $ show (Impl (Not to) (Impl (Impl from to) (Not (Not to))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl from to) (Not (Not to)))) (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Not (Not to))) (Not (Impl from to)))) (Impl (Not to) (Not (Impl from to)))))
    putStrLn $ show (Impl (Impl (Not to) (Impl (Impl (Impl from to) (Not (Not to))) (Not (Impl from to)))) (Impl (Not to) (Not (Impl from to))))
    putStrLn $ show (Impl (Not to) (Not (Impl from to)))
    putStrLn $ show (Impl (Not to) (Not (Not (Impl from to))))
    putStrLn $ show (Impl (Impl (Not to) (Not (Not (Impl from to)))) (Not (Not to)))
    putStrLn $ show (Not (Not to))

loop []        context !implMap !allAsMap = return ()
loop (expr:xs) context !implMap !allAsMap =
                               let
                               exprAsLineProof = min (maybeModusPonens expr implMap allAsMap) (min (maybeAxiom expr) (maybeHypothesis context expr))
                               index = Map.size allAsMap
                                   in
                                         let
                                         newAllAsMap = if isNothing (Map.lookup expr allAsMap) then (Map.insert expr index allAsMap) else allAsMap
                                         newImplMap = if isImpl expr then Map.insertWith (++) (getTo expr) [(getFrom expr)] implMap else implMap
                                             in
                                                 do
                                                     case exprAsLineProof of
                                                      (Axiom 10 (Impl (Not (Not a)) a')) -> writeAxiom10 a
                                                      (Axiom _ e) -> writeAxiomAndHypothesis e
                                                      (Hypothesis _ e) -> writeAxiomAndHypothesis e
                                                      (ModusPonens from to) -> writeModusPonens from to
                                                     loop xs context newImplMap newAllAsMap

-- Main
main = do
    firstLine  <- getLine
    let problem  = parseProblem $ alexScanTokens firstLine
    let context = (reverse $ getContext problem)
    let contextMap  = Map.fromList (zip context [1..])
    let result   = getResult problem
    proof'    <- getContents
    let proof   = map (parseExpression . alexScanTokens) (filter  (/=[]) (lines proof'))
    putStrLn (showProblem context (Not (Not result)))
    loop proof contextMap Map.empty Map.empty